@@ -382,7 +382,7 @@ module Puma
               else
                 # if ThreadPool out_of_band code is running, we don't want to add
                 # clients until the code is finished.
-                sleep 0.001 while pool.out_of_band_running
+                pool.wait_while_out_of_band_running
 
                 # only use delay when clustered and busy
                 if pool.busy_threads >= @max_threads
@@ -508,7 +508,7 @@ module Puma
           next_request_ready = if client.has_back_to_back_requests?
             with_force_shutdown(client) { client.process_back_to_back_requests }
           else
-            nil
+            with_force_shutdown(client) { client.eagerly_finish }
           end
 
           if next_request_ready
@@ -685,13 +685,13 @@ module Puma
       stats = @thread_pool&.stats || {}
       stats[:max_threads]    = @max_threads
       stats[:requests_count] = @requests_count
-      stats[:reactor_max] = @reactor.reactor_max
+      stats[:reactor_max] = @reactor.reactor_max if @reactor
       reset_max
       stats
     end
 
     def reset_max
-      @reactor.reactor_max = 0
+      @reactor.reactor_max = 0 if @reactor
       @thread_pool.reset_max
     end
 
