@@ -26,8 +26,10 @@ class TestIntegration < PumaTest
 
   def setup
     @server = nil
-    @config_file = nil
     @server_log = +''
+    @server_stopped = false
+    @config_file = nil
+
     @pid = nil
     @ios_to_close = []
     @bind_path    = tmp_path('.sock')
@@ -38,7 +40,8 @@ class TestIntegration < PumaTest
   def teardown
     if @server && @control_tcp_port && Puma.windows?
       cli_pumactl 'stop'
-    elsif @server && @pid && !Puma.windows?
+    # don't close if we've already done so
+    elsif @server && @pid && !@server_stopped && !Puma.windows?
       stop_server @pid, signal: :INT
     end
 
@@ -137,6 +140,7 @@ class TestIntegration < PumaTest
   # that is already stopped/killed, especially since Process.wait2 is
   # blocking
   def stop_server(pid = @pid, signal: :TERM)
+    @server_stopped = true
     begin
       Process.kill signal, pid
     rescue Errno::ESRCH
@@ -227,7 +231,7 @@ class TestIntegration < PumaTest
       end
     rescue StandardError => e
       error_retries += 1
-      raise(e, "Waiting for server to log #{match_obj.inspect}") if error_retries == LOG_ERROR_QTY
+      raise(Minitest::Assertion,  "Waiting for server to log #{match_obj.inspect} raised #{e.class}") if error_retries == LOG_ERROR_QTY
       sleep LOG_ERROR_SLEEP
       retry
     end
@@ -460,7 +464,7 @@ class TestIntegration < PumaTest
             else
               mutex.synchronize { replies[:unexpected_response] += 1 }
             end
-          rescue Errno::ECONNRESET, Errno::EBADF, Errno::ENOTCONN
+          rescue Errno::ECONNRESET, Errno::EBADF, Errno::ENOTCONN, Errno::ENOTSOCK
             # connection was accepted but then closed
             # client would see an empty response
             # Errno::EBADF Windows may not be able to make a connection
@@ -512,7 +516,7 @@ class TestIntegration < PumaTest
           begin
             get_worker_pids phase, log: log
             # Wait with an exponential backoff before signaling next restart
-            sleep 0.15 * (2 ** restart_count)
+            sleep 0.15 * restart_count
           rescue Minitest::Assertion # Timeout
             run = false
           rescue Errno::EBADF # bad restart?
