@@ -1,12 +1,15 @@
+# frozen_string_literal: true
+
 # Copyright (c) 2011 Evan Phoenix
 # Copyright (c) 2005 Zed A. Shaw
 
 require_relative "helper"
+require_relative "helpers/integration"
 require "digest"
 
 require "puma/puma_http11"
 
-class Http11ParserTest < Minitest::Test
+class Http11ParserTest < TestIntegration
 
   parallelize_me!
 
@@ -22,7 +25,7 @@ class Http11ParserTest < Minitest::Test
     assert nread == parser.nread, "Number read returned from execute does not match"
 
     assert_equal '/', req['REQUEST_PATH']
-    assert_equal 'HTTP/1.1', req['HTTP_VERSION']
+    assert_equal 'HTTP/1.1', req['SERVER_PROTOCOL']
     assert_equal '/?a=1', req['REQUEST_URI']
     assert_equal 'GET', req['REQUEST_METHOD']
     assert_nil req['FRAGMENT']
@@ -63,7 +66,7 @@ class Http11ParserTest < Minitest::Test
 
     assert_equal "GET", req['REQUEST_METHOD']
     assert_equal 'http://192.168.1.96:3000/api/v1/matches/test?1=1', req['REQUEST_URI']
-    assert_equal 'HTTP/1.1', req['HTTP_VERSION']
+    assert_equal 'HTTP/1.1', req['SERVER_PROTOCOL']
 
     assert_nil req['REQUEST_PATH']
     assert_nil req['FRAGMENT']
@@ -114,7 +117,6 @@ class Http11ParserTest < Minitest::Test
   end
 
   def test_semicolon_in_path
-    skip_if :jruby # Not yet supported on JRuby, see https://github.com/puma/puma/issues/1978
     parser = Puma::HttpParser.new
     req = {}
     get = "GET /forums/1/path;stillpath/2375?page=1 HTTP/1.1\r\n\r\n"
@@ -132,14 +134,58 @@ class Http11ParserTest < Minitest::Test
     res = count.to_s + "/"
 
     if readable
-      res << Digest::SHA1.hexdigest(rand(count * 100).to_s) * (count / 40)
+      res << Digest(:SHA1).hexdigest(rand(count * 100).to_s) * (count / 40)
     else
-      res << Digest::SHA1.digest(rand(count * 100).to_s) * (count / 20)
+      res << Digest(:SHA1).digest(rand(count * 100).to_s) * (count / 20)
     end
 
     res
   end
 
+  def test_get_const_length
+    skip_unless :jruby
+
+    envs = %w[PUMA_REQUEST_URI_MAX_LENGTH PUMA_REQUEST_PATH_MAX_LENGTH PUMA_QUERY_STRING_MAX_LENGTH]
+    default_exp = [1024 * 12, 8192, 10 * 1024]
+    tests = [{ envs: %w[60000 61000 62000], exp: [60000, 61000, 62000], error_indexes: [] },
+             { envs: ['', 'abc', nil], exp: default_exp, error_indexes: [1] },
+             { envs: %w[-4000 0 3000.45], exp: default_exp, error_indexes: [0, 1, 2] }]
+    cli_config = <<~CONFIG
+        app do |_|
+          require 'json'
+          [200, {}, [{ MAX_REQUEST_URI_LENGTH: org.jruby.puma.Http11::MAX_REQUEST_URI_LENGTH,
+                       MAX_REQUEST_PATH_LENGTH: org.jruby.puma.Http11::MAX_REQUEST_PATH_LENGTH,
+                       MAX_QUERY_STRING_LENGTH: org.jruby.puma.Http11::MAX_QUERY_STRING_LENGTH,
+                       MAX_REQUEST_URI_LENGTH_ERR: org.jruby.puma.Http11::MAX_REQUEST_URI_LENGTH_ERR,
+                       MAX_REQUEST_PATH_LENGTH_ERR: org.jruby.puma.Http11::MAX_REQUEST_PATH_LENGTH_ERR,
+                       MAX_QUERY_STRING_LENGTH_ERR: org.jruby.puma.Http11::MAX_QUERY_STRING_LENGTH_ERR }.to_json]]
+        end
+    CONFIG
+
+    tests.each do |conf|
+      cli_server 'test/rackup/hello.ru',
+                      env: {envs[0]  => conf[:envs][0], envs[1] => conf[:envs][1], envs[2] => conf[:envs][2]},
+                      merge_err: true,
+                      config: cli_config
+      result = JSON.parse read_body(connect)
+
+      assert_equal conf[:exp][0], result['MAX_REQUEST_URI_LENGTH']
+      assert_equal conf[:exp][1], result['MAX_REQUEST_PATH_LENGTH']
+      assert_equal conf[:exp][2], result['MAX_QUERY_STRING_LENGTH']
+
+      assert_includes result['MAX_REQUEST_URI_LENGTH_ERR'], "longer than the #{conf[:exp][0]} allowed length"
+      assert_includes result['MAX_REQUEST_PATH_LENGTH_ERR'], "longer than the #{conf[:exp][1]} allowed length"
+      assert_includes result['MAX_QUERY_STRING_LENGTH_ERR'], "longer than the #{conf[:exp][2]} allowed length"
+
+      conf[:error_indexes].each do |index|
+        assert_includes @server_log, "The value #{conf[:envs][index]} for #{envs[index]} is invalid. "\
+          "Using default value #{default_exp[index]} instead"
+      end
+
+      stop_server
+     end
+  end
+
   def test_max_uri_path_length
     parser = Puma::HttpParser.new
     req = {}
