@@ -163,8 +163,6 @@ class TestPumaServer < PumaTest
   end
 
   def test_pipe_body_http11
-    skip if ::Puma::IS_WINDOWS && RUBY_VERSION < '2.7' #encoding issue?
-
     random_bytes = SecureRandom.random_bytes(4096)
 
     r, w = IO.pipe(binmode: true)
@@ -185,8 +183,6 @@ class TestPumaServer < PumaTest
   end
 
   def test_pipe_body_http10
-    skip if ::Puma::IS_WINDOWS && RUBY_VERSION < '2.7' #encoding issue?
-
     bytes = 4_096
     random_bytes = SecureRandom.random_bytes(bytes)
 
@@ -207,12 +203,61 @@ class TestPumaServer < PumaTest
     r&.close
   end
 
+  # Older Rubies may only read 16k, maybe OS dependent
+  def test_request_body_small
+    data = nil
+
+    server_run do |env|
+      data = env['rack.input'].read
+      [200, {}, [env['rack.input'].to_s]]
+    end
+
+    small_body = 'x' * 15 * 1_024
+
+    small_body_bytes = small_body.bytesize
+
+    socket = send_http "PUT / HTTP/1.0\r\n" \
+      "Content-Length: #{small_body_bytes}\r\n\r\n" \
+      "#{small_body}"
+
+    assert_includes socket.read_response, 'StringIO'
+
+    assert_equal small_body_bytes, data.bytesize
+    assert_equal small_body, data
+    # below intermitttently fails on GitHub Actions, may work locally
+    # assert_equal 0, @server.stats[:reactor_max]
+  end
+
+  def test_request_body_large
+    data = nil
+
+    server_run do |env|
+      data = env['rack.input'].read
+      [200, {}, [env['rack.input'].to_s]]
+    end
+
+    large_body = 'x' * 70 * 1_024
+
+    large_body_bytes = large_body.bytesize
+
+    socket = send_http "PUT / HTTP/1.0\r\n" \
+      "Content-Length: #{large_body_bytes}\r\n\r\n" \
+      "#{large_body}"
+
+    assert_includes socket.read_response, 'StringIO'
+
+    assert_equal large_body_bytes, data.bytesize
+    assert_equal large_body, data
+    # below intermitttently fails on GitHub Actions, may work locally
+    # assert_equal 0, @server.stats[:reactor_max]
+  end
+
   def test_proper_stringio_body
     data = nil
 
     server_run do |env|
       data = env['rack.input'].read
-      [200, {}, ["ok"]]
+      [200, {}, [env['rack.input'].to_s]]
     end
 
     fifteen = "1" * 15
@@ -222,7 +267,7 @@ class TestPumaServer < PumaTest
     sleep 0.1 # important so that the previous data is sent as a packet
     socket << fifteen
 
-    socket.read_response
+    assert_includes socket.read_response, 'StringIO'
 
     assert_equal "#{fifteen}#{fifteen}", data
   end
@@ -306,7 +351,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response "HEAD / HTTP/1.0\r\n\r\n"
 
-    assert_equal "HTTP/1.0 200 OK\r\nFoo: Bar\r\nContent-Length: 5\r\n\r\n", response
+    assert_equal "HTTP/1.0 200 OK\r\nfoo: Bar\r\ncontent-length: 5\r\n\r\n", response
   end
 
   def test_GET_with_empty_body_has_sane_chunking
@@ -314,7 +359,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response "HEAD / HTTP/1.0\r\n\r\n"
 
-    assert_equal "HTTP/1.0 200 OK\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.0 200 OK\r\ncontent-length: 0\r\n\r\n", response
   end
 
   def test_back_to_back_no_content
@@ -330,8 +375,8 @@ class TestPumaServer < PumaTest
     )
 
     assert_equal(
-      "HTTP/1.1 200 OK\r\nContent-Length: 4\r\n\r\nok 1" \
-      "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length: 4\r\n\r\nok 2", data
+      "HTTP/1.1 200 OK\r\ncontent-length: 4\r\n\r\nok 1" \
+      "HTTP/1.1 200 OK\r\nconnection: close\r\ncontent-length: 4\r\n\r\nok 2", data
     )
     assert_equal ["", ""], bodies
   end
@@ -349,8 +394,8 @@ class TestPumaServer < PumaTest
     )
 
     assert_equal(
-      "HTTP/1.1 200 OK\r\nContent-Length: 4\r\n\r\nok 1" \
-      "HTTP/1.1 200 OK\r\nContent-Length: 4\r\n\r\nok 2", data
+      "HTTP/1.1 200 OK\r\ncontent-length: 4\r\n\r\nok 1" \
+      "HTTP/1.1 200 OK\r\ncontent-length: 4\r\n\r\nok 2", data
     )
     assert_equal ["a", ""], bodies
   end
@@ -370,15 +415,15 @@ class TestPumaServer < PumaTest
     data = socket.sysread 1_024
 
     assert_equal(
-      "HTTP/1.1 200 OK\r\nContent-Length: 5\r\n\r\nhello" \
-      "HTTP/1.1 200 OK\r\nContent-Length: 7\r\n\r\ngoodbye", data
+      "HTTP/1.1 200 OK\r\ncontent-length: 5\r\n\r\nhello" \
+      "HTTP/1.1 200 OK\r\ncontent-length: 7\r\n\r\ngoodbye", data
     )
 
     socket << "GET / HTTP/1.1\r\nTransfer-Encoding: chunked\r\n\r\n1\r\nH\r\n4\r\nello\r\n0\r\n\r\n"
 
     response = socket.read_response
 
-    assert_equal "HTTP/1.1 200 OK\r\nContent-Length: 5\r\n\r\nHello", response
+    assert_equal "HTTP/1.1 200 OK\r\ncontent-length: 5\r\n\r\nHello", response
   end
 
   def test_early_hints_works
@@ -387,7 +432,8 @@ class TestPumaServer < PumaTest
      [200, { "X-Hello" => "World" }, ["Hello world!"]]
     end
 
-    response = send_http_read_response "HEAD / HTTP/1.0\r\n\r\n"
+    # two requests must be read
+    response = send_http_read_all "HEAD / HTTP/1.0\r\n\r\n"
 
     expected_resp = <<~EOF.gsub("\n", "\r\n") + "\r\n"
       HTTP/1.1 103 Early Hints
@@ -395,8 +441,8 @@ class TestPumaServer < PumaTest
       Link: </script.js>; rel=preload
 
       HTTP/1.0 200 OK
-      X-Hello: World
-      Content-Length: 12
+      x-hello: World
+      content-length: 12
     EOF
 
     assert_equal true, @server.early_hints
@@ -434,8 +480,8 @@ class TestPumaServer < PumaTest
 
     expected_resp = <<~EOF.gsub("\n", "\r\n") + "\r\n"
       HTTP/1.0 200 OK
-      X-Hello: World
-      Content-Length: 12
+      x-hello: World
+      content-length: 12
     EOF
 
     assert_nil @server.early_hints
@@ -464,7 +510,7 @@ class TestPumaServer < PumaTest
 
     # Content Too Large
     assert_equal "HTTP/1.1 413 #{STATUS_CODES[413]}", response.status
-    assert_equal ["Content-Length: 17"], response.headers
+    assert_equal ["content-length: 17"], response.headers
   end
 
   def test_GET_with_no_body_has_sane_chunking
@@ -472,7 +518,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response "HEAD / HTTP/1.0\r\n\r\n"
 
-    assert_equal "HTTP/1.0 200 OK\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.0 200 OK\r\ncontent-length: 0\r\n\r\n", response
   end
 
   def test_doesnt_print_backtrace_in_production
@@ -504,7 +550,7 @@ class TestPumaServer < PumaTest
     response = send_http_read_response GET_10
 
     assert_equal 'HTTP/1.0 500 Internal Server Error', response.status
-    assert_match(/Content-Type: application\/json/, response)
+    assert_match(/content-type: application\/json/, response)
     assert_match(/{}\n$/, response)
   end
 
@@ -613,7 +659,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response GET_10
 
-    assert_equal "HTTP/1.0 449 CUSTOM\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.0 449 CUSTOM\r\ncontent-length: 0\r\n\r\n", response
   end
 
   def test_custom_http_codes_11
@@ -621,7 +667,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response "GET / HTTP/1.1\r\nConnection: close\r\n\r\n"
 
-    assert_equal "HTTP/1.1 449 CUSTOM\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 449 CUSTOM\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
   end
 
   def test_HEAD_returns_content_headers
@@ -630,7 +676,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response "HEAD / HTTP/1.0\r\n\r\n"
 
-    assert_equal "HTTP/1.0 200 OK\r\nContent-Type: application/pdf\r\nContent-Length: 4242\r\n\r\n", response
+    assert_equal "HTTP/1.0 200 OK\r\ncontent-type: application/pdf\r\ncontent-length: 4242\r\n\r\n", response
   end
 
   def test_status_hook_fires_when_server_changes_states
@@ -826,7 +872,7 @@ class TestPumaServer < PumaTest
     req  = "GET / HTTP/1.1\r\nConnection: Keep-Alive\r\n\r\n"
     response = send_http_read_response req
 
-    assert_equal ["Content-Type: plain/text", "Content-Length: 6"], response.headers
+    assert_equal ["content-type: plain/text", "content-length: 6"], response.headers
     assert_equal "hello\n", response.body
   end
 
@@ -835,7 +881,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response "GET / HTTP/1.1\r\nConnection: close\r\n\r\n"
 
-    assert_equal "HTTP/1.1 200 OK\r\nContent-Type: plain/text\r\nConnection: close\r\nContent-Length: 5\r\n\r\nhello", response
+    assert_equal "HTTP/1.1 200 OK\r\ncontent-type: plain/text\r\nconnection: close\r\ncontent-length: 5\r\n\r\nhello", response
   end
 
   def test_http_11_keep_alive_without_body
@@ -855,7 +901,7 @@ class TestPumaServer < PumaTest
 
     # No Content
     assert_equal "HTTP/1.1 204 #{STATUS_CODES[204]}", response.status
-    assert_equal ["Connection: close"], response.headers
+    assert_equal ["connection: close"], response.headers
   end
 
   def test_http_11_enable_keep_alives_by_default
@@ -864,8 +910,8 @@ class TestPumaServer < PumaTest
     req  = "GET / HTTP/1.1\r\n\r\n"
     response = send_http_read_response req
 
-    # No "Connection: close" header.
-    assert_equal ["Content-Type: plain/text", "Content-Length: 6"], response.headers
+    # No "connection: close" header.
+    assert_equal ["content-type: plain/text", "content-length: 6"], response.headers
     assert_equal "hello\n", response.body
   end
   def test_http_11_enable_keep_alives_true
@@ -874,8 +920,8 @@ class TestPumaServer < PumaTest
     req  = "GET / HTTP/1.1\r\n\r\n"
     response = send_http_read_response req
 
-    # No "Connection: close" header.
-    assert_equal ["Content-Type: plain/text", "Content-Length: 6"], response.headers
+    # No "connection: close" header.
+    assert_equal ["content-type: plain/text", "content-length: 6"], response.headers
     assert_equal "hello\n", response.body
   end
 
@@ -885,8 +931,8 @@ class TestPumaServer < PumaTest
     req  = "GET / HTTP/1.1\r\n\r\n"
     response = send_http_read_response req
 
-    # Assert the "Connection: close" header is present with keep-alives disabled.
-    assert_equal ["Content-Type: plain/text", "Connection: close", "Content-Length: 6"], response.headers
+    # Assert the "connection: close" header is present with keep-alives disabled.
+    assert_equal ["content-type: plain/text", "connection: close", "content-length: 6"], response.headers
     assert_equal "hello\n", response.body
   end
 
@@ -898,7 +944,7 @@ class TestPumaServer < PumaTest
     response = send_http_read_response req
 
     assert_equal "HTTP/1.0 200 OK", response.status
-    assert_equal ["Content-Type: plain/text", "Connection: Keep-Alive", "Content-Length: 6"],
+    assert_equal ["content-type: plain/text", "connection: keep-alive", "content-length: 6"],
       response.headers
     assert_equal "hello\n", response.body
   end
@@ -908,7 +954,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response "GET / HTTP/1.0\r\nConnection: close\r\n\r\n"
 
-    assert_equal "HTTP/1.0 200 OK\r\nContent-Type: plain/text\r\nContent-Length: 5\r\n\r\nhello", response
+    assert_equal "HTTP/1.0 200 OK\r\ncontent-type: plain/text\r\ncontent-length: 5\r\n\r\nhello", response
   end
 
   def test_http_10_keep_alive_without_body
@@ -916,7 +962,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response "GET / HTTP/1.0\r\nConnection: Keep-Alive\r\n\r\n"
 
-    assert_equal "HTTP/1.0 204 No Content\r\nConnection: Keep-Alive\r\n\r\n", response
+    assert_equal "HTTP/1.0 204 No Content\r\nconnection: keep-alive\r\n\r\n", response
   end
 
   def test_http_10_close_without_body
@@ -930,9 +976,10 @@ class TestPumaServer < PumaTest
   def test_Expect_100
     server_run { [200, {}, [""]] }
 
-    response = send_http_read_response "GET / HTTP/1.1\r\nConnection: close\r\nExpect: 100-continue\r\n\r\n"
+    # two requests must be read
+    response = send_http_read_all "GET / HTTP/1.1\r\nConnection: close\r\nExpect: 100-continue\r\n\r\n"
 
-    assert_equal "HTTP/1.1 100 Continue\r\n\r\nHTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 100 Continue\r\n\r\nHTTP/1.1 200 OK\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
   end
 
   def test_chunked_request
@@ -948,7 +995,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response "GET / HTTP/1.1\r\nConnection: close\r\nTransfer-Encoding: gzip,chunked\r\n\r\n1\r\nh\r\n4\r\nello\r\n0\r\n\r\n"
 
-    assert_equal "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
     assert_equal "hello", body
     assert_equal "5", content_length
     assert_nil transfer_encoding
@@ -976,7 +1023,7 @@ class TestPumaServer < PumaTest
     response = skt.read_response
     path1 = req_body_path
 
-    assert_equal "HTTP/1.1 200 OK\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\ncontent-length: 0\r\n\r\n", response
     assert_equal "hello", body
     assert_equal "5", content_length
     assert_nil transfer_encoding
@@ -986,7 +1033,7 @@ class TestPumaServer < PumaTest
     path2 = req_body_path
 
     # same as above
-    assert_equal "HTTP/1.1 200 OK\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\ncontent-length: 0\r\n\r\n", response
     assert_equal "hello", body
     assert_equal "5", content_length
     assert_nil transfer_encoding
@@ -1018,7 +1065,7 @@ class TestPumaServer < PumaTest
 
       response = send_http_read_response request
 
-      assert_equal "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+      assert_equal "HTTP/1.1 200 OK\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
       assert_equal size, Integer(content_length)
       assert_equal request_body, body
     end
@@ -1037,7 +1084,7 @@ class TestPumaServer < PumaTest
     header = "GET / HTTP/1.1\r\nConnection: close\r\nContent-Length: 200\r\nTransfer-Encoding: chunked\r\n\r\n"
     response = send_http_read_response "#{header}1;t=#{'x' * (max_chunk_header_size + 2)}\r\n1\r\nh\r\n4\r\nello\r\n0\r\n\r\n"
 
-    assert_equal "HTTP/1.1 400 Bad Request\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 400 Bad Request\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
   end
 
   def test_chunked_request_invalid_extension_header_length_split
@@ -1066,7 +1113,7 @@ class TestPumaServer < PumaTest
 
       response = socket.read_response
       refute_equal 'hello', body
-      assert_equal "HTTP/1.1 400 Bad Request\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+      assert_equal "HTTP/1.1 400 Bad Request\r\nConnection: close\r\ncontent-length: 0\r\n\r\n", response
 
     # errors raised vary by OS
     rescue Errno::EPIPE, Errno::ECONNABORTED, Errno::ECONNRESET
@@ -1090,7 +1137,7 @@ class TestPumaServer < PumaTest
 
     response = socket.read_response
 
-    assert_equal "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
     assert_equal "hello", body
     assert_equal "5", content_length
   end
@@ -1111,7 +1158,7 @@ class TestPumaServer < PumaTest
 
     response = socket.read_response
 
-    assert_equal "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
     assert_equal "hello", body
     assert_equal "5", content_length
   end
@@ -1132,7 +1179,7 @@ class TestPumaServer < PumaTest
 
     response = socket.read_response
 
-    assert_equal "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
     assert_equal "hello", body
     assert_equal "5", content_length
   end
@@ -1153,7 +1200,7 @@ class TestPumaServer < PumaTest
 
     response = socket.read_response
 
-    assert_equal "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
     assert_equal "hello", body
     assert_equal "5", content_length
   end
@@ -1174,7 +1221,7 @@ class TestPumaServer < PumaTest
 
     response = socket.read_response
 
-    assert_equal "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
     assert_equal "hello", body
     assert_equal "5", content_length
   end
@@ -1204,7 +1251,7 @@ class TestPumaServer < PumaTest
 
     response = socket.read_response
 
-    assert_equal "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
     assert_equal (part1 + 'b'), body
     assert_equal "4201", content_length
   end
@@ -1226,7 +1273,7 @@ class TestPumaServer < PumaTest
 
     response = socket.read_response
 
-    assert_equal "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
     assert_equal 'hello', body
     assert_equal "5", content_length
   end
@@ -1248,7 +1295,7 @@ class TestPumaServer < PumaTest
 
     response = socket.read_response
 
-    assert_equal "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
     assert_equal 'hello', body
     assert_equal "5", content_length
   end
@@ -1283,7 +1330,7 @@ class TestPumaServer < PumaTest
     socket << data2
 
     response = socket.read_response
-    assert_equal "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
     assert_equal 9*1_024, body.bytesize
     assert_equal 9*1_024, content_length.to_i
     assert_equal '012345678', body.delete('x')
@@ -1300,7 +1347,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response "GET / HTTP/1.1\r\nConnection: close\r\nTransfer-Encoding: Chunked\r\n\r\n1\r\nh\r\n4\r\nello\r\n0\r\n\r\n"
 
-    assert_equal "HTTP/1.1 200 OK\r\nConnection: close\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\nconnection: close\r\ncontent-length: 0\r\n\r\n", response
     assert_equal "hello", body
     assert_equal "5", content_length
   end
@@ -1316,7 +1363,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response "GET / HTTP/1.1\r\nConnection: Keep-Alive\r\nTransfer-Encoding: chunked\r\n\r\n1\r\nh\r\n4\r\nello\r\n0\r\n\r\n"
 
-    assert_equal "HTTP/1.1 200 OK\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\ncontent-length: 0\r\n\r\n", response
     assert_equal "hello", body
     assert_equal "5", content_length
   end
@@ -1343,7 +1390,7 @@ class TestPumaServer < PumaTest
 
     response = socket.read_response
 
-    assert_equal "HTTP/1.1 200 OK\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\ncontent-length: 0\r\n\r\n", response
     assert_equal "hello", body
     assert_equal "5", content_length
     sleep 0.05 if TRUFFLE
@@ -1356,11 +1403,61 @@ class TestPumaServer < PumaTest
 
     response = socket.read_response
 
-    assert_equal "HTTP/1.1 200 OK\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\ncontent-length: 0\r\n\r\n", response
     assert_equal "goodbye", body
     assert_equal "7", content_length
   end
 
+  def test_chunked_keep_alive_twenty_back_to_back
+    req_count = 20
+    requests = 0
+
+    server_run(max_keep_alive: 21) { |env|
+      requests += 1
+      env['rack.input'].read
+      [200, {}, ["Request_#{requests}"]]
+    }
+
+    req = "GET / HTTP/1.1\r\nX-Forwarded-For: 127.0.0.1\r\nConnection: Keep-Alive\r\nTransfer-Encoding: chunked\r\n\r\n1\r\nh\r\n4\r\nello\r\n0\r\n\r\n"
+
+    expected = String.new # rubocop: disable Performance/UnfreezeString
+    req_count.times do |i|
+      expected << "HTTP/1.1 200 OK\r\ncontent-length: #{(i+1).to_s.length + 8}\r\n\r\nRequest_#{i+1}"
+    end
+
+    responses = send_http_read_all(req * req_count)
+
+    assert_equal req_count, responses.scan('Request_').length
+
+    assert_equal expected, responses
+  end
+
+  def test_close_socket_after_max_keepalive_requests
+    max_keep_alive = 2
+    requests = 0
+    response = ''
+
+    server_run(max_keep_alive: max_keep_alive) { |env|
+      requests += 1
+      [200, {}, [format('Request_%02d', requests)]]
+    }
+
+    socket = new_socket
+
+    # EOFError - Ubuntu, Errno::ECONNRESET - macOS, Errno::ECONNABORTED - windows
+    assert_raises EOFError, Errno::ECONNABORTED, Errno::ECONNRESET do
+      (max_keep_alive + 1).times do |i|
+        socket << GET_11
+        response = socket.read_response
+        refute_includes(Puma::Const::CONNECTION_CLOSE, response) if i < max_keep_alive + 1
+      end
+    end
+    sleep 0.25 # seems to be needed for macOS
+    assert_raises(Errno::EPIPE) { socket << GET_11 }
+    assert_includes response, format('Request_%02d', max_keep_alive)
+    assert_includes response, Puma::Const::CONNECTION_CLOSE
+  end
+
   def test_chunked_keep_alive_two_back_to_back_with_set_remote_address
     body = nil
     content_length = nil
@@ -1376,7 +1473,7 @@ class TestPumaServer < PumaTest
 
     response = socket.read_response
 
-    assert_equal "HTTP/1.1 200 OK\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\ncontent-length: 0\r\n\r\n", response
     assert_equal "hello", body
     assert_equal "5", content_length
     assert_equal "127.0.0.1", remote_addr
@@ -1386,7 +1483,7 @@ class TestPumaServer < PumaTest
 
     response = socket.read_response
 
-    assert_equal "HTTP/1.1 200 OK\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 200 OK\r\ncontent-length: 0\r\n\r\n", response
     assert_equal "goodbye", body
     assert_equal "7", content_length
     assert_equal "127.0.0.2", remote_addr
@@ -1425,7 +1522,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response "HEAD / HTTP/1.0\r\n\r\n"
 
-    assert_equal "HTTP/1.0 200 OK\r\nX-Empty-Header: \r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.0 200 OK\r\nx-empty-header: \r\ncontent-length: 0\r\n\r\n", response
   end
 
   def test_request_body_wait
@@ -1467,7 +1564,7 @@ class TestPumaServer < PumaTest
 
   def test_open_connection_wait(**options)
     server_run(**options) { [200, {}, ["Hello"]] }
-    socket = send_http nil
+    socket = new_socket
     sleep 0.1
     socket << GET_10
     assert_equal 'Hello', socket.read_body
@@ -1483,7 +1580,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response "HEAD / HTTP/1.0\r\n\r\n"
 
-    assert_match "X-header: first line\r\nX-header: second line\r\n", response
+    assert_match "x-header: first line\r\nx-header: second line\r\n", response
   end
 
   def test_newline_splits_in_early_hint
@@ -1549,6 +1646,7 @@ class TestPumaServer < PumaTest
   # There are three different tests because there are three ways to set header
   # content in Puma. Regular (rack env), early hints, and a special case for
   # overriding content-length.
+
   {"cr" => "\r", "lf" => "\n", "crlf" => "\r\n"}.each do |suffix, line_ending|
     # The cr-only case for the following test was CVE-2020-5247
     define_method(:"test_prevent_response_splitting_headers_#{suffix}") do
@@ -1654,23 +1752,23 @@ class TestPumaServer < PumaTest
     response = socket.read_response
 
     assert_equal "HTTP/1.1 200 OK", response.status
-    assert_equal ["Content-Length: 0"], response.headers
+    assert_equal ["content-length: 0"], response.headers
 
     socket << "GET / HTTP/1.1\r\nConnection: close\r\n\r\n"
     response = socket.read_response
 
     assert_equal "HTTP/1.1 200 OK", response.status
-    assert_equal ["Connection: close", "Content-Length: 0"], response.headers
+    assert_equal ["connection: close", "content-length: 0"], response.headers
 
     socket = send_http "GET / HTTP/1.1\r\n\r\n"
     response = socket.read_response
     assert_equal "HTTP/1.1 200 OK", response.status
-    assert_equal ["Content-Length: 0"], response.headers
+    assert_equal ["content-length: 0"], response.headers
 
     socket << "GET / HTTP/1.1\r\nConnection: close\r\n\r\n"
     response = socket.read_response
     assert_equal "HTTP/1.1 200 OK", response.status
-    assert_equal ["Connection: close", "Content-Length: 0"], response.headers
+    assert_equal ["connection: close", "content-length: 0"], response.headers
   end
 
   def test_http10_connection_header_queue
@@ -1680,26 +1778,26 @@ class TestPumaServer < PumaTest
     response = socket.read_response
 
     assert_equal "HTTP/1.0 200 OK", response.status
-    assert_equal ["Connection: Keep-Alive", "Content-Length: 0"], response.headers
+    assert_equal ["connection: keep-alive", "content-length: 0"], response.headers
 
     socket << "GET / HTTP/1.0\r\n\r\n"
     response = socket.read_response
     assert_equal "HTTP/1.0 200 OK", response.status
-    assert_equal ["Content-Length: 0"], response.headers
+    assert_equal ["content-length: 0"], response.headers
   end
 
   def test_http11_connection_header_no_queue
     server_run(queue_requests: false) { [200, {}, [""]] }
     response = send_http_read_response GET_11
     assert_equal "HTTP/1.1 200 OK", response.status
-    assert_equal ["Connection: close", "Content-Length: 0"], response.headers
+    assert_equal ["connection: close", "content-length: 0"], response.headers
   end
 
   def test_http10_connection_header_no_queue
     server_run(queue_requests: false) { [200, {}, [""]] }
     response = send_http_read_response GET_10
     assert_equal "HTTP/1.0 200 OK", response.status
-    assert_equal ["Content-Length: 0"], response.headers
+    assert_equal ["content-length: 0"], response.headers
   end
 
   def stub_accept_nonblock(error)
@@ -1720,6 +1818,7 @@ class TestPumaServer < PumaTest
   # System-resource errors such as EMFILE should not be silently swallowed by accept loop.
   def test_accept_emfile
     stub_accept_nonblock Errno::EMFILE.new('accept(2)')
+    sleep 0.1 if Puma::IS_WINDOWS && Puma::IS_ARM
     refute_empty @log_writer.stderr.string, "Expected EMFILE error not logged"
   end
 
@@ -1917,7 +2016,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response GET_11
     # Not Found
-    assert_equal "HTTP/1.1 404 #{STATUS_CODES[404]}\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 404 #{STATUS_CODES[404]}\r\ncontent-length: 0\r\n\r\n", response
   end
 
   def test_empty_body_array_no_content_length
@@ -1925,7 +2024,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response GET_11
     # Not Found
-    assert_equal "HTTP/1.1 404 #{STATUS_CODES[404]}\r\nContent-Length: 0\r\n\r\n", response
+    assert_equal "HTTP/1.1 404 #{STATUS_CODES[404]}\r\ncontent-length: 0\r\n\r\n", response
   end
 
   def test_empty_body_enum
@@ -1933,7 +2032,7 @@ class TestPumaServer < PumaTest
 
     response = send_http_read_response GET_11
     # Not Found
-    assert_equal "HTTP/1.1 404 #{STATUS_CODES[404]}\r\nTransfer-Encoding: chunked\r\n\r\n0\r\n\r\n", response
+    assert_equal "HTTP/1.1 404 #{STATUS_CODES[404]}\r\ntransfer-encoding: chunked\r\n\r\n0\r\n\r\n", response
   end
 
   def test_form_data_encoding_windows_bom
