@@ -99,8 +99,12 @@ class TestPumaServerSSL < PumaTest
   end
 
   def test_very_large_return
+    # This test frequntly fails on Darwin TruffleRuby, 512k was used chosen
+    # becuase 1mb also failed
+    # failure is OpenSSL::SSL::SSLError: SSL_read: record layer failure
+    body_size = Puma::IS_OSX && TRUFFLE ? 512 * 1_024 : 2_056_610
     start_server
-    giant = "x" * 2056610
+    giant = "x" * body_size
 
     @server.app = proc { [200, {}, [giant]] }
 
